@@ -1,59 +1,31 @@
-/**************************************************************************
-**
-** This file is part of Qt Creator
-**
-** Copyright (c) 2011 Nokia Corporation and/or its subsidiary(-ies).
-**
-** Contact: Nokia Corporation (info@qt.nokia.com)
-**
-**
-** GNU Lesser General Public License Usage
-**
-** This file may be used under the terms of the GNU Lesser General Public
-** License version 2.1 as published by the Free Software Foundation and
-** appearing in the file LICENSE.LGPL included in the packaging of this file.
-** Please review the following information to ensure the GNU Lesser General
-** Public License version 2.1 requirements will be met:
-** http://www.gnu.org/licenses/old-licenses/lgpl-2.1.html.
-**
-** In addition, as a special exception, Nokia gives you certain additional
-** rights. These rights are described in the Nokia Qt LGPL Exception
-** version 1.1, included in the file LGPL_EXCEPTION.txt in this package.
-**
-** Other Usage
-**
-** Alternatively, this file may be used in accordance with the terms and
-** conditions contained in a signed written agreement between you and Nokia.
-**
-** If you have questions regarding the use of this file, please contact
-** Nokia at qt-info@nokia.com.
-**
-**************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 
 #include <QtTest>
 #include <QObject>
 
-#include <AST.h>
-#include <ASTVisitor.h>
-#include <TranslationUnit.h>
-#include <CppDocument.h>
-#include <Literals.h>
-#include <LookupContext.h>
-#include <Name.h>
-#include <ResolveExpression.h>
-#include <Symbols.h>
-#include <Overview.h>
+#include <cplusplus/AST.h>
+#include <cplusplus/ASTVisitor.h>
+#include <cplusplus/CppDocument.h>
+#include <cplusplus/Literals.h>
+#include <cplusplus/LookupContext.h>
+#include <cplusplus/Name.h>
+#include <cplusplus/NamePrettyPrinter.h>
+#include <cplusplus/Overview.h>
+#include <cplusplus/ResolveExpression.h>
+#include <cplusplus/Symbols.h>
+#include <cplusplus/TranslationUnit.h>
 
 //TESTED_COMPONENT=src/libs/cplusplus
 using namespace CPlusPlus;
 
-template <template <typename, typename> class _Map, typename _T1, typename _T2>
-_Map<_T2, _T1> invert(const _Map<_T1, _T2> &m)
+template<typename _T1, typename _T2>
+QMultiMap<_T2, _T1> invert(const QMap<_T1, _T2> &m)
 {
-    _Map<_T2, _T1> i;
-    typename _Map<_T1, _T2>::const_iterator it = m.constBegin();
+    QMultiMap<_T2, _T1> i;
+    typename QMap<_T1, _T2>::const_iterator it = m.constBegin();
     for (; it != m.constEnd(); ++it) {
-        i.insertMulti(it.value(), it.key());
+        i.insert(it.value(), it.key());
     }
     return i;
 }
@@ -88,9 +60,12 @@ class tst_Lookup: public QObject
 {
     Q_OBJECT
 
-private Q_SLOTS:
+private slots:
     void base_class_defined_1();
 
+    void document_functionAt_data();
+    void document_functionAt();
+
     // Objective-C
     void simple_class_1();
     void class_with_baseclass();
@@ -103,6 +78,9 @@ private Q_SLOTS:
     void templates_3();
     void templates_4();
     void templates_5();
+
+    void minimalname_data();
+    void minimalname();
 };
 
 void tst_Lookup::base_class_defined_1()
@@ -114,12 +92,12 @@ void tst_Lookup::base_class_defined_1()
         "class derived: public base {};\n";
 
     Document::Ptr doc = Document::create("base_class_defined_1");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
     QVERIFY(doc->diagnosticMessages().isEmpty());
-    QCOMPARE(doc->globalSymbolCount(), 2U);
+    QCOMPARE(doc->globalSymbolCount(), 2);
 
     Snapshot snapshot;
     snapshot.insert(doc);
@@ -149,11 +127,92 @@ void tst_Lookup::base_class_defined_1()
 
     QCOMPARE(classSymbols.size(), 2);
 
-    const QMap<Class *, ClassSpecifierAST *> classToAST =
-            invert(classSymbols.asMap());
+    const QMultiMap<Class *, ClassSpecifierAST *> classToAST = invert(classSymbols.asMap());
 
-    QVERIFY(classToAST.value(baseClass) != 0);
-    QVERIFY(classToAST.value(derivedClass) != 0);
+    QVERIFY(!classToAST.values(baseClass).isEmpty());
+    QVERIFY(!classToAST.values(derivedClass).isEmpty());
+}
+
+void tst_Lookup::document_functionAt_data()
+{
+    QTest::addColumn<QByteArray>("source");
+    QTest::addColumn<int>("line");
+    QTest::addColumn<int>("column");
+    QTest::addColumn<QString>("expectedFunction");
+    QTest::addColumn<int>("expectedOpeningDeclaratorParenthesisLine");
+    QTest::addColumn<int>("expectedClosingBraceLine");
+
+    QByteArray source = "\n"
+            "void Foo::Bar() {\n" // line 1
+            "    \n"
+            "    for (int i=0; i < 10; ++i) {\n" // line 3
+            "        \n"
+            "    }\n" // line 5
+            "}\n";
+    QString expectedFunction = QString::fromLatin1("Foo::Bar");
+    QTest::newRow("nonInline1") << source << 1 << 2 << QString() << -1 << -1;
+    QTest::newRow("nonInline2") << source << 1 << 11 << expectedFunction << 1 << 6;
+    QTest::newRow("nonInline3") << source << 2 << 2 << expectedFunction << 1 << 6;
+    QTest::newRow("nonInline4") << source << 3 << 10 << expectedFunction << 1 << 6;
+    QTest::newRow("nonInline5") << source << 4 << 3 << expectedFunction << 1 << 6;
+    QTest::newRow("nonInline6") << source << 6 << 1 << expectedFunction << 1 << 6;
+
+    source = "\n"
+            "namespace N {\n" // line 1
+            "class C {\n"
+            "    void f()\n" // line 3
+            "    {\n"
+            "    }\n" // line 5
+            "};\n"
+            "}\n"; // line 7
+    expectedFunction = QString::fromLatin1("N::C::f");
+    QTest::newRow("inline1") << source << 1 << 2 << QString() << -1 << -1;
+    QTest::newRow("inline2") << source << 2 << 10 << QString() << -1 << -1;
+    QTest::newRow("inline2") << source << 3 << 10 << expectedFunction << 3 << 5;
+
+    source = "\n"
+            "void f(Helper helper = [](){})\n" // line 1
+            "{\n"
+            "}\n"; // line 3
+    expectedFunction = QString::fromLatin1("f");
+    QTest::newRow("inlineWithLambdaArg1") << source << 2 << 1 << expectedFunction << 1 << 3;
+
+    source = "\n"
+             "int g_global = 456;\n" // line 1
+             "int foo()\n"  // line 2
+             "{\n"
+             "  g_global = 89;\n" // line 4
+             "  std::function<int(int)> cb = [](int) { return 1; };\n"
+             "  g_global = 222;\n"
+             "}\n"; // line 7
+    expectedFunction = QString::fromLatin1("foo");
+    QTest::newRow("inlineWithLambda1") << source << 4 << 1 << expectedFunction << 2 << 7;
+    QTest::newRow("inlineWithLambda2") << source << 6 << 1 << expectedFunction << 2 << 7;
+}
+
+void tst_Lookup::document_functionAt()
+{
+    QFETCH(QByteArray, source);
+    QFETCH(int, line);
+    QFETCH(int, column);
+    QFETCH(QString, expectedFunction);
+    QFETCH(int, expectedOpeningDeclaratorParenthesisLine);
+    QFETCH(int, expectedClosingBraceLine);
+
+    Document::Ptr doc = Document::create("document_functionAt");
+    doc->setUtf8Source(source);
+    doc->parse();
+    doc->check();
+    QVERIFY(doc->diagnosticMessages().isEmpty());
+
+    int actualOpeningDeclaratorParenthesisLine = -1;
+    int actualClosingBraceLine = -1;
+    const QString actualFunction = doc->functionAt(line, column,
+                                                   &actualOpeningDeclaratorParenthesisLine,
+                                                   &actualClosingBraceLine);
+    QCOMPARE(actualFunction, expectedFunction);
+    QCOMPARE(actualOpeningDeclaratorParenthesisLine, expectedOpeningDeclaratorParenthesisLine);
+    QCOMPARE(actualClosingBraceLine, expectedClosingBraceLine);
 }
 
 void tst_Lookup::simple_class_1()
@@ -163,12 +222,12 @@ void tst_Lookup::simple_class_1()
         "@implementation Zoo +(id)alloc{} -(id)init{} -(void)dealloc{} @end\n";
 
     Document::Ptr doc = Document::create("simple_class_1");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
     QVERIFY(doc->diagnosticMessages().isEmpty());
-    QCOMPARE(doc->globalSymbolCount(), 2U);
+    QCOMPARE(doc->globalSymbolCount(), 2);
 
     Snapshot snapshot;
     snapshot.insert(doc);
@@ -176,12 +235,12 @@ void tst_Lookup::simple_class_1()
     ObjCClass *iface = doc->globalSymbolAt(0)->asObjCClass();
     QVERIFY(iface);
     QVERIFY(iface->isInterface());
-    QCOMPARE(iface->memberCount(), 2U);
+    QCOMPARE(iface->memberCount(), 2);
 
     ObjCClass *impl = doc->globalSymbolAt(1)->asObjCClass();
     QVERIFY(impl);
     QVERIFY(!impl->isInterface());
-    QCOMPARE(impl->memberCount(), 3U);
+    QCOMPARE(impl->memberCount(), 3);
 
     Declaration *allocMethodIface = iface->memberAt(0)->asDeclaration();
     QVERIFY(allocMethodIface);
@@ -226,12 +285,12 @@ void tst_Lookup::class_with_baseclass()
                               "@implementation Zoo +(id)alloc{} -(id)init{} -(void)dealloc{} @end\n";
 
     Document::Ptr doc = Document::create("class_with_baseclass");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
     QVERIFY(doc->diagnosticMessages().isEmpty());
-    QCOMPARE(doc->globalSymbolCount(), 3U);
+    QCOMPARE(doc->globalSymbolCount(), 3);
 
     Snapshot snapshot;
     snapshot.insert(doc);
@@ -241,7 +300,7 @@ void tst_Lookup::class_with_baseclass()
     ObjCClass *baseZoo = doc->globalSymbolAt(0)->asObjCClass();
     QVERIFY(baseZoo);
     QVERIFY(!baseZoo->isInterface());
-    QCOMPARE(baseZoo->memberCount(), 2U);
+    QCOMPARE(baseZoo->memberCount(), 2);
 
     ObjCClass *zooIface = doc->globalSymbolAt(1)->asObjCClass();
     QVERIFY(zooIface);
@@ -251,7 +310,7 @@ void tst_Lookup::class_with_baseclass()
     ObjCClass *zooImpl = doc->globalSymbolAt(2)->asObjCClass();
     QVERIFY(zooImpl);
     QVERIFY(!zooImpl->isInterface());
-    QCOMPARE(zooImpl->memberCount(), 3U);
+    QCOMPARE(zooImpl->memberCount(), 3);
 
     Declaration *baseDecl = baseZoo->memberAt(0)->asDeclaration();
     QVERIFY(baseDecl);
@@ -287,20 +346,20 @@ void tst_Lookup::class_with_protocol_with_protocol()
                               "@implementation Zoo +(id)alloc{} -(id)init{} -(void)dealloc{} @end\n";
 
     Document::Ptr doc = Document::create("class_with_protocol_with_protocol");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
     QVERIFY(doc->diagnosticMessages().isEmpty());
-    QCOMPARE(doc->globalSymbolCount(), 4U);
+    QCOMPARE(doc->globalSymbolCount(), 4);
 
     Snapshot snapshot;
     snapshot.insert(doc);
 
     ObjCProtocol *P1 = doc->globalSymbolAt(0)->asObjCProtocol();
     QVERIFY(P1);
-    QCOMPARE(P1->memberCount(), 1U);
-    QCOMPARE(P1->protocolCount(), 0U);
+    QCOMPARE(P1->memberCount(), 1);
+    QCOMPARE(P1->protocolCount(), 0);
 
     Declaration *p1method = P1->memberAt(0)->asDeclaration();
     QVERIFY(p1method);
@@ -308,8 +367,8 @@ void tst_Lookup::class_with_protocol_with_protocol()
 
     ObjCProtocol *P2 = doc->globalSymbolAt(1)->asObjCProtocol();
     QVERIFY(P2);
-    QCOMPARE(P2->memberCount(), 1U);
-    QCOMPARE(P2->protocolCount(), 1U);
+    QCOMPARE(P2->memberCount(), 1);
+    QCOMPARE(P2->protocolCount(), 1);
     QCOMPARE(QLatin1String(P2->protocolAt(0)->name()->identifier()->chars()), QLatin1String("P1"));
 
     ObjCClass *zooImpl = doc->globalSymbolAt(3)->asObjCClass();
@@ -341,12 +400,12 @@ void tst_Lookup::iface_impl_scoping()
                               "@implementation Scooping-(int)method1:(int)arg{return arg;}@end\n";
 
     Document::Ptr doc = Document::create("class_with_protocol_with_protocol");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
     QVERIFY(doc->diagnosticMessages().isEmpty());
-    QCOMPARE(doc->globalSymbolCount(), 2U);
+    QCOMPARE(doc->globalSymbolCount(), 2);
 
     Snapshot snapshot;
     snapshot.insert(doc);
@@ -358,19 +417,19 @@ void tst_Lookup::iface_impl_scoping()
     QVERIFY(impl);
     QVERIFY(!impl->isInterface());
 
-    QCOMPARE(iface->memberCount(), 2U);
-    QCOMPARE(impl->memberCount(), 1U);
+    QCOMPARE(iface->memberCount(), 2);
+    QCOMPARE(impl->memberCount(), 1);
 
     ObjCMethod *method1Impl = impl->memberAt(0)->asObjCMethod();
     QVERIFY(method1Impl);
     QCOMPARE(method1Impl->identifier()->chars(), "method1");
 
     // get the body of method1
-    QCOMPARE(method1Impl->memberCount(), 2U);
+    QCOMPARE(method1Impl->memberCount(), 2);
     Argument *method1Arg = method1Impl->memberAt(0)->asArgument();
     QVERIFY(method1Arg);
     QCOMPARE(method1Arg->identifier()->chars(), "arg");
-    QVERIFY(method1Arg->type()->isIntegerType());
+    QVERIFY(method1Arg->type()->asIntegerType());
 
     Block *method1Body = method1Impl->memberAt(1)->asBlock();
     QVERIFY(method1Body);
@@ -378,13 +437,13 @@ void tst_Lookup::iface_impl_scoping()
     const LookupContext context(doc, snapshot);
 
     { // verify if we can resolve "arg" in the body
-        QCOMPARE(method1Impl->argumentCount(), 1U);
+        QCOMPARE(method1Impl->argumentCount(), 1);
         Argument *arg = method1Impl->argumentAt(0)->asArgument();
         QVERIFY(arg);
         QVERIFY(arg->name());
         QVERIFY(arg->name()->identifier());
         QCOMPARE(arg->name()->identifier()->chars(), "arg");
-        QVERIFY(arg->type()->isIntegerType());
+        QVERIFY(arg->type()->asIntegerType());
 
         const QList<LookupItem> candidates = context.lookup(arg->name(), method1Body->enclosingScope());
         QCOMPARE(candidates.size(), 1);
@@ -431,7 +490,7 @@ void tst_Lookup::templates_1()
             "    l.end(); // std::_List_iterator<Point>\n"
             "}\n";
     Document::Ptr doc = Document::create("templates_1");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
@@ -462,7 +521,7 @@ void tst_Lookup::templates_2()
             "}\n"
 ;
     Document::Ptr doc = Document::create("templates_2");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
@@ -487,7 +546,7 @@ void tst_Lookup::templates_3()
             "    l.at(0); // const Point &\n"
             "}\n";
     Document::Ptr doc = Document::create("templates_3");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
@@ -528,7 +587,7 @@ void tst_Lookup::templates_4()
             "    (*l); // Point &\n"
             "}\n";
     Document::Ptr doc = Document::create("templates_4");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
@@ -558,12 +617,54 @@ void tst_Lookup::templates_5()
             "    a.get(); // const Point &\n"
             "}\n";
     Document::Ptr doc = Document::create("templates_5");
-    doc->setSource(source);
+    doc->setUtf8Source(source);
     doc->parse();
     doc->check();
 
     QVERIFY(doc->diagnosticMessages().isEmpty());
 }
 
+void tst_Lookup::minimalname_data()
+{
+    QTest::addColumn<QByteArray>("source");
+    QTest::addColumn<int>("index");
+
+    QTest::newRow("inlineNamespace1")
+            << QByteArray("namespace std { inline namespace __cxx11 { class string{}; } }\n")
+            << 0;
+
+    // This case is extracted from libstdc++ 5.4.0.
+    // The inline namespace is re-opened as non-inline, which is not standard
+    // compliant. However, gcc does this and clang only issues a warning.
+    QTest::newRow("inlineNamespace2")
+            << QByteArray("namespace std { inline namespace __cxx11 {} }\n"
+                          "namespace std { namespace __cxx11 { class string{}; } }\n")
+            << 1;
+}
+
+void tst_Lookup::minimalname()
+{
+    QFETCH(QByteArray, source);
+    QFETCH(int, index);
+
+    Document::Ptr doc = Document::create("minimalname");
+    doc->setUtf8Source(source);
+    doc->parse();
+    doc->check();
+
+    Snapshot snapshot;
+    snapshot.insert(doc);
+    LookupContext ctx(doc, snapshot);
+    Control control;
+    Symbol *symbol = doc->globalSymbolAt(unsigned(index))
+            ->asNamespace()->memberAt(0)->asNamespace()->memberAt(0);
+
+    const Name *minimalName = LookupContext::minimalName(symbol, ctx.globalNamespace(), &control);
+
+    Overview oo;
+    const QString minimalNameAsString = NamePrettyPrinter(&oo)(minimalName);
+    QCOMPARE(minimalNameAsString, QString::fromUtf8("std::string"));
+}
+
 QTEST_APPLESS_MAIN(tst_Lookup)
 #include "tst_lookup.moc"
